from __future__ import annotations

import asyncio
from typing import cast

import pytest
from arp_llm.providers.dev_mock import DevMockChatModel
from arp_standard_model import (
    CompositeBeginRequest,
    CompositeExecutorCancelCompositeNodeRunParams,
    CompositeExecutorCancelCompositeNodeRunRequest,
    CompositeExecutorHealthRequest,
    CompositeExecutorVersionRequest,
    EndpointLocator,
    NodeRunCompleteRequest,
    NodeRunTerminalState,
    NodeTypeRef,
)
from arp_standard_server import ArpServerError

import jarvis_composite_executor.executor as executor_module
from jarvis_composite_executor.config import CompositeConfig
from jarvis_composite_executor.executor import CompositeExecutor


class FakeRunCoordinator:
    def __init__(self) -> None:
        self.completed: list[NodeRunCompleteRequest] = []

    async def create_node_runs(self, body):
        _ = body
        raise RuntimeError("not implemented")

    async def get_node_run(self, node_run_id: str):
        _ = node_run_id
        raise RuntimeError("not implemented")

    async def report_node_run_evaluation(self, node_run_id: str, body) -> None:
        _ = node_run_id, body
        raise RuntimeError("not implemented")

    async def complete_node_run(self, node_run_id: str, body) -> None:
        _ = node_run_id
        self.completed.append(body)


class DummySelection:
    async def generate_candidate_set(self, body):
        _ = body
        raise RuntimeError("not implemented")


class DummyRegistry:
    async def get_node_type(self, node_type_id: str, version: str | None = None):
        _ = node_type_id, version
        raise RuntimeError("not implemented")


class DummyTask:
    def __init__(self) -> None:
        self.canceled = False

    def cancel(self) -> None:
        self.canceled = True


def _config() -> CompositeConfig:
    return CompositeConfig(
        max_steps_default=1,
        max_depth_default=1,
        poll_timeout_sec=0.01,
        poll_interval_sec=0.0,
        arggen_max_retries=0,
        context_window=None,
        max_concurrency=1,
    )


def _request(node_run_id: str = "node") -> CompositeBeginRequest:
    return CompositeBeginRequest(
        run_id="run",
        node_run_id=node_run_id,
        node_type_ref=NodeTypeRef(node_type_id="jarvis.composite.planner", version="0.3.7"),
        inputs={"goal": "test"},
        coordinator_endpoint=EndpointLocator.model_validate("http://127.0.0.1:8081"),
    )


def test_health_and_version() -> None:
    executor = CompositeExecutor(
        selection_service=DummySelection(),
        node_registry=DummyRegistry(),
        llm=DevMockChatModel(),
        config=_config(),
        service_name="svc",
        service_version="1.2.3",
    )

    health = asyncio.run(executor.health(CompositeExecutorHealthRequest()))
    version = asyncio.run(executor.version(CompositeExecutorVersionRequest()))

    assert health.status.value == "ok"
    assert version.service_name == "svc"


def test_cancel_composite_node_run_cancels_task() -> None:
    executor = CompositeExecutor(
        selection_service=DummySelection(),
        node_registry=DummyRegistry(),
        llm=DevMockChatModel(),
        config=_config(),
    )
    task = DummyTask()
    executor._tasks["node"] = cast(asyncio.Task[None], task)

    asyncio.run(
        executor.cancel_composite_node_run(
            CompositeExecutorCancelCompositeNodeRunRequest(
                params=CompositeExecutorCancelCompositeNodeRunParams(node_run_id="node")
            )
        )
    )

    assert "node" in executor._canceled
    assert task.canceled is True


def test_start_assignment_when_already_running() -> None:
    executor = CompositeExecutor(selection_service=None, node_registry=None, llm=None, config=None)
    executor._tasks["node"] = cast(asyncio.Task[None], DummyTask())

    executor._start_assignment(_request())

    assert executor._tasks["node"] is not None


def test_start_assignment_outside_loop() -> None:
    executor = CompositeExecutor(selection_service=None, node_registry=None, llm=None, config=None)

    executor._start_assignment(_request())

    assert "node" not in executor._tasks


def test_drive_assignment_handles_errors(monkeypatch: pytest.MonkeyPatch) -> None:
    class FakeDriver:
        def __init__(self, **kwargs) -> None:
            _ = kwargs

        async def run(self, request) -> None:
            _ = request
            raise ArpServerError(code="boom", message="bad", status_code=500)

    coordinator = FakeRunCoordinator()

    monkeypatch.setattr(executor_module, "CompositeAssignmentDriver", FakeDriver)

    executor = CompositeExecutor(
        selection_service=DummySelection(),
        node_registry=DummyRegistry(),
        llm=DevMockChatModel(),
        config=_config(),
        run_coordinator_factory=lambda _: coordinator,
    )

    asyncio.run(executor._drive_assignment(_request()))

    assert coordinator.completed[0].state == NodeRunTerminalState.failed


def test_drive_assignment_handles_cancelled(monkeypatch: pytest.MonkeyPatch) -> None:
    class FakeDriver:
        def __init__(self, **kwargs) -> None:
            _ = kwargs

        async def run(self, request) -> None:
            _ = request
            raise asyncio.CancelledError

    coordinator = FakeRunCoordinator()

    monkeypatch.setattr(executor_module, "CompositeAssignmentDriver", FakeDriver)

    executor = CompositeExecutor(
        selection_service=DummySelection(),
        node_registry=DummyRegistry(),
        llm=DevMockChatModel(),
        config=_config(),
        run_coordinator_factory=lambda _: coordinator,
    )

    asyncio.run(executor._drive_assignment(_request()))

    assert coordinator.completed[0].state == NodeRunTerminalState.canceled


def test_drive_assignment_handles_generic_exception(monkeypatch: pytest.MonkeyPatch) -> None:
    class FakeDriver:
        def __init__(self, **kwargs) -> None:
            _ = kwargs

        async def run(self, request) -> None:
            _ = request
            raise RuntimeError("boom")

    coordinator = FakeRunCoordinator()

    monkeypatch.setattr(executor_module, "CompositeAssignmentDriver", FakeDriver)

    executor = CompositeExecutor(
        selection_service=DummySelection(),
        node_registry=DummyRegistry(),
        llm=DevMockChatModel(),
        config=_config(),
        run_coordinator_factory=lambda _: coordinator,
    )

    asyncio.run(executor._drive_assignment(_request()))

    error = coordinator.completed[0].error
    assert error is not None
    assert error.code == "composite_error"


def test_ensure_ready_missing_dependencies() -> None:
    executor = CompositeExecutor(selection_service=DummySelection(), node_registry=None, llm=None, config=None)
    with pytest.raises(ArpServerError) as excinfo:
        executor._ensure_ready()
    assert excinfo.value.code == "node_registry_missing"

    executor = CompositeExecutor(selection_service=DummySelection(), node_registry=DummyRegistry(), llm=None, config=None)
    with pytest.raises(ArpServerError) as excinfo:
        executor._ensure_ready()
    assert excinfo.value.code == "composite_llm_missing"

    executor = CompositeExecutor(selection_service=DummySelection(), node_registry=DummyRegistry(), llm=DevMockChatModel(), config=None)
    with pytest.raises(ArpServerError) as excinfo:
        executor._ensure_ready()
    assert excinfo.value.code == "composite_config_missing"



from __future__ import annotations

from typing import Any

from arp_standard_model import Candidate, CandidateSet, NodeTypeRef


def choose_candidate(candidate_set: CandidateSet) -> Candidate | None:
    if not candidate_set.candidates:
        return None

    selection = _extensions_dict(candidate_set.extensions)
    if selection.get("jarvis.selection.needs_planner") is True:
        if (planner_ref := _planner_ref(selection)) is not None:
            for candidate in candidate_set.candidates:
                if candidate.node_type_ref == planner_ref:
                    return candidate
            return None

        for candidate in candidate_set.candidates:
            if candidate.node_type_ref.node_type_id.startswith("jarvis.composite.planner"):
                return candidate
        return None

    return candidate_set.candidates[0]


def _extensions_dict(extensions: Any) -> dict[str, Any]:
    if extensions is None:
        return {}
    if hasattr(extensions, "model_dump"):
        return extensions.model_dump(exclude_none=True)
    if isinstance(extensions, dict):
        return dict(extensions)
    return {}


def _planner_ref(selection: dict[str, Any]) -> NodeTypeRef | None:
    raw = selection.get("jarvis.selection.planner_node_type_ref")
    if not isinstance(raw, dict):
        return None
    node_type_id = raw.get("node_type_id")
    version = raw.get("version")
    if not isinstance(node_type_id, str) or not isinstance(version, str):
        return None
    return NodeTypeRef(node_type_id=node_type_id, version=version)



from __future__ import annotations

import argparse
import os

import uvicorn

LOG_LEVEL = os.getenv("LOG_LEVEL", "INFO").upper()

LOG_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "format": "%(asctime)s %(levelname)s %(name)s: %(message)s",
        },
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "default",
            "stream": "ext://sys.stdout",
        },
    },
    "root": {
        "handlers": ["console"],
        "level": LOG_LEVEL,
    },
    "loggers": {
        "uvicorn": {"handlers": ["console"], "level": LOG_LEVEL, "propagate": False},
        "uvicorn.error": {"level": LOG_LEVEL},
        "uvicorn.access": {"handlers": ["console"], "level": LOG_LEVEL, "propagate": False},
    },
}


def main() -> None:
    parser = argparse.ArgumentParser(description="Start the JARVIS Composite Executor server.")
    parser.add_argument("--host", default="127.0.0.1")
    parser.add_argument("--port", type=int, default=8083)
    parser.add_argument("--reload", action="store_true", help="Enable auto-reload (dev only).")
    args = parser.parse_args()

    if args.reload:
        uvicorn.run(
            "jarvis_composite_executor.app:app",
            host=args.host,
            port=args.port,
            reload=True,
            log_config=LOG_CONFIG,
        )
        return

    from .app import create_app

    app = create_app()
    uvicorn.run(app, host=args.host, port=args.port, reload=False, log_config=LOG_CONFIG)


if __name__ == "__main__":
    main()



from __future__ import annotations

import pytest
from arp_standard_model import NodeTypeRef

from jarvis_composite_executor import schema_utils


def test_normalize_schema_requires_all_properties() -> None:
    schema = {
        "type": "object",
        "properties": {
            "text": {"type": "string"},
            "limit": {"type": "integer"},
        },
        "required": ["text"],
        "additionalProperties": False,
    }

    normalized = schema_utils.normalize_schema_for_llm(schema)

    assert sorted(normalized["required"]) == ["limit", "text"]
    assert normalized["properties"]["limit"]["type"] == ["integer", "null"]
    assert normalized["properties"]["text"]["type"] == "string"


def test_normalize_schema_handles_anyof() -> None:
    schema = {
        "anyOf": [
            {"type": "string"},
            {
                "type": "object",
                "properties": {"value": {"type": "string"}},
                "required": ["value"],
            },
        ]
    }

    normalized = schema_utils.normalize_schema_for_llm(schema)

    assert isinstance(normalized["anyOf"], list)
    assert normalized["anyOf"][1]["properties"]["value"]["type"] == "string"


def test_optional_anyof_allows_null() -> None:
    schema = {
        "type": "object",
        "properties": {
            "maybe": {"anyOf": [{"type": "string"}]},
        },
        "required": [],
    }

    normalized = schema_utils.normalize_schema_for_llm(schema)

    any_of = normalized["properties"]["maybe"]["anyOf"]
    assert {"type": "null"} in any_of


def test_normalize_schema_handles_oneof_and_arrays() -> None:
    schema = {
        "oneOf": [
            {"type": "string"},
            {"type": "array", "items": {"type": "integer"}},
        ]
    }

    normalized = schema_utils.normalize_schema_for_llm(schema)

    assert normalized["oneOf"][1]["items"]["type"] == "integer"


def test_optional_type_list_allows_null() -> None:
    schema = {
        "type": "object",
        "properties": {"maybe": {"type": ["string", "number"]}},
        "required": [],
    }

    normalized = schema_utils.normalize_schema_for_llm(schema)

    assert "null" in normalized["properties"]["maybe"]["type"]


def test_optional_empty_schema_adds_anyof_null() -> None:
    schema = {
        "type": "object",
        "properties": {"maybe": {}},
        "required": [],
    }

    normalized = schema_utils.normalize_schema_for_llm(schema)

    assert normalized["properties"]["maybe"]["anyOf"][1] == {"type": "null"}


def test_validate_json_schema() -> None:
    schema = {"type": "object", "properties": {"text": {"type": "string"}}, "required": ["text"]}
    schema_utils.validate_json_schema({"text": "ok"}, schema)
    with pytest.raises(Exception):
        schema_utils.validate_json_schema({"text": 123}, schema)


def test_as_dict_model_dump() -> None:
    ref = NodeTypeRef(node_type_id="node", version="1")
    assert schema_utils._as_dict(ref) == {"node_type_id": "node", "version": "1"}



from __future__ import annotations

import asyncio

import pytest
from arp_llm.providers.dev_mock import DevMockChatFixture, DevMockChatModel
from arp_standard_model import (
    Candidate,
    CandidateSet,
    CompositeBeginRequest,
    EndpointLocator,
    EvaluationStatus,
    NodeKind,
    NodeRun,
    NodeRunCompleteRequest,
    NodeRunEvaluationReportRequest,
    NodeRunState,
    NodeRunTerminalState,
    NodeRunsCreateRequest,
    NodeRunsCreateResponse,
    NodeType,
    NodeTypeRef,
)
from arp_standard_server import ArpServerError

from jarvis_composite_executor.config import CompositeConfig
from jarvis_composite_executor.engine.arggen import ArgGen
from jarvis_composite_executor.engine.driver import CompositeAssignmentDriver
from jarvis_composite_executor.engine.planner import Planner


class RecordingRunCoordinator:
    def __init__(self) -> None:
        self.completed: list[NodeRunCompleteRequest] = []
        self.evaluations: list[NodeRunEvaluationReportRequest] = []

    async def create_node_runs(self, body: NodeRunsCreateRequest) -> NodeRunsCreateResponse:
        _ = body
        raise RuntimeError("not implemented")

    async def get_node_run(self, node_run_id: str) -> NodeRun:
        _ = node_run_id
        raise RuntimeError("not implemented")

    async def complete_node_run(self, node_run_id: str, body: NodeRunCompleteRequest) -> None:
        _ = node_run_id
        self.completed.append(body)

    async def report_node_run_evaluation(self, node_run_id: str, body: NodeRunEvaluationReportRequest) -> None:
        _ = node_run_id
        self.evaluations.append(body)


class EmptySelection:
    async def generate_candidate_set(self, body) -> CandidateSet:
        _ = body
        return CandidateSet(candidate_set_id="set", subtask_id="sub", candidates=[])


class CandidateSelection:
    async def generate_candidate_set(self, body) -> CandidateSet:
        _ = body
        return CandidateSet(
            candidate_set_id="set",
            subtask_id="sub",
            candidates=[Candidate(node_type_ref=NodeTypeRef(node_type_id="node", version="1"), score=1.0)],
        )


class EmptyNodeRegistry:
    async def get_node_type(self, node_type_id: str, version: str | None = None) -> NodeType:
        _ = node_type_id, version
        return NodeType(
            node_type_id="node",
            version="1",
            kind=NodeKind.atomic,
            description=None,
            input_schema={"type": "object", "properties": {"text": {"type": "string"}}, "required": ["text"]},
        )


def _base_config() -> CompositeConfig:
    return CompositeConfig(
        max_steps_default=2,
        max_depth_default=2,
        poll_timeout_sec=0.05,
        poll_interval_sec=0.0,
        arggen_max_retries=0,
        context_window=None,
        max_concurrency=None,
    )


def _base_request(node_run_id: str = "node") -> CompositeBeginRequest:
    return CompositeBeginRequest(
        run_id="run",
        node_run_id=node_run_id,
        node_type_ref=NodeTypeRef(node_type_id="jarvis.composite.planner", version="0.3.7"),
        inputs={"goal": "test"},
        coordinator_endpoint=EndpointLocator.model_validate("http://127.0.0.1:8081"),
    )


def test_driver_canceled_before_start() -> None:
    run_coordinator = RecordingRunCoordinator()
    planner = Planner(DevMockChatModel(fixtures=[DevMockChatFixture(text="p", parsed={"subtasks": [], "summary": None})]))
    arggen = ArgGen(DevMockChatModel(fixtures=[DevMockChatFixture(text="a", parsed={})]), max_retries=0)
    driver = CompositeAssignmentDriver(
        run_coordinator=run_coordinator,
        selection=CandidateSelection(),
        node_registry=EmptyNodeRegistry(),
        planner=planner,
        arggen=arggen,
        config=_base_config(),
        canceled={"node"},
    )

    asyncio.run(driver.run(_base_request()))

    assert run_coordinator.completed[0].state == NodeRunTerminalState.canceled


def test_driver_empty_plan_finalizes() -> None:
    run_coordinator = RecordingRunCoordinator()
    planner_fixture = DevMockChatFixture(text="planner", parsed={"subtasks": [], "summary": "done"})
    planner = Planner(DevMockChatModel(fixtures=[planner_fixture]))
    arggen = ArgGen(DevMockChatModel(fixtures=[DevMockChatFixture(text="a", parsed={})]), max_retries=0)

    driver = CompositeAssignmentDriver(
        run_coordinator=run_coordinator,
        selection=CandidateSelection(),
        node_registry=EmptyNodeRegistry(),
        planner=planner,
        arggen=arggen,
        config=_base_config(),
        canceled=set(),
    )

    asyncio.run(driver.run(_base_request()))

    assert run_coordinator.evaluations[0].evaluation_result.status == EvaluationStatus.success
    assert run_coordinator.completed[0].state == NodeRunTerminalState.succeeded


def test_driver_no_candidates() -> None:
    planner_fixture = DevMockChatFixture(
        text="planner",
        parsed={"subtasks": [{"subtask_id": "s1", "goal": "do thing", "notes": None}], "summary": None},
    )
    planner = Planner(DevMockChatModel(fixtures=[planner_fixture]))
    arggen = ArgGen(DevMockChatModel(fixtures=[DevMockChatFixture(text="a", parsed={})]), max_retries=0)

    driver = CompositeAssignmentDriver(
        run_coordinator=RecordingRunCoordinator(),
        selection=EmptySelection(),
        node_registry=EmptyNodeRegistry(),
        planner=planner,
        arggen=arggen,
        config=_base_config(),
        canceled=set(),
    )

    with pytest.raises(ArpServerError) as excinfo:
        asyncio.run(driver.run(_base_request()))

    assert excinfo.value.code == "composite_no_candidates"


def test_driver_no_child_runs() -> None:
    class NoChildRunCoordinator(RecordingRunCoordinator):
        async def create_node_runs(self, body: NodeRunsCreateRequest) -> NodeRunsCreateResponse:
            _ = body
            return NodeRunsCreateResponse(node_runs=[])

        async def get_node_run(self, node_run_id: str) -> NodeRun:
            raise RuntimeError("unexpected")

    planner_fixture = DevMockChatFixture(
        text="planner",
        parsed={"subtasks": [{"subtask_id": "s1", "goal": "do thing", "notes": None}], "summary": None},
    )
    arggen_fixture = DevMockChatFixture(text="arggen", parsed={"text": "ok"})
    llm = DevMockChatModel(fixtures=[planner_fixture, arggen_fixture])

    driver = CompositeAssignmentDriver(
        run_coordinator=NoChildRunCoordinator(),
        selection=CandidateSelection(),
        node_registry=EmptyNodeRegistry(),
        planner=Planner(llm),
        arggen=ArgGen(llm, max_retries=0),
        config=_base_config(),
        canceled=set(),
    )

    with pytest.raises(ArpServerError) as excinfo:
        asyncio.run(driver.run(_base_request()))

    assert excinfo.value.code == "composite_no_child_runs"


def test_driver_failed_child_breaks() -> None:
    class FailedRunCoordinator(RecordingRunCoordinator):
        async def create_node_runs(self, body: NodeRunsCreateRequest) -> NodeRunsCreateResponse:
            _ = body
            node_run = NodeRun(
                node_run_id="child",
                node_type_ref=NodeTypeRef(node_type_id="node", version="1"),
                state=NodeRunState.failed,
                run_id="run",
            )
            return NodeRunsCreateResponse(node_runs=[node_run])

        async def get_node_run(self, node_run_id: str) -> NodeRun:
            _ = node_run_id
            return NodeRun(
                node_run_id="child",
                node_type_ref=NodeTypeRef(node_type_id="node", version="1"),
                state=NodeRunState.failed,
                run_id="run",
            )

    planner_fixture = DevMockChatFixture(
        text="planner",
        parsed={"subtasks": [{"subtask_id": "s1", "goal": "do thing", "notes": None}], "summary": None},
    )
    arggen_fixture = DevMockChatFixture(text="arggen", parsed={"text": "ok"})
    llm = DevMockChatModel(fixtures=[planner_fixture, arggen_fixture])

    coordinator = FailedRunCoordinator()
    driver = CompositeAssignmentDriver(
        run_coordinator=coordinator,
        selection=CandidateSelection(),
        node_registry=EmptyNodeRegistry(),
        planner=Planner(llm),
        arggen=ArgGen(llm, max_retries=0),
        config=_base_config(),
        canceled=set(),
    )

    asyncio.run(driver.run(_base_request()))

    assert driver._canceled == set()
    assert coordinator.completed[0].state == NodeRunTerminalState.failed


def test_driver_composite_inputs_are_filled() -> None:
    class CompositeNodeRegistry(EmptyNodeRegistry):
        async def get_node_type(self, node_type_id: str, version: str | None = None) -> NodeType:
            _ = node_type_id, version
            return NodeType(
                node_type_id="planner",
                version="1",
                kind=NodeKind.composite,
                description=None,
                input_schema={"type": "object", "properties": {}, "required": []},
            )

    class CapturingRunCoordinator(RecordingRunCoordinator):
        def __init__(self) -> None:
            super().__init__()
            self.created: list[NodeRunsCreateRequest] = []

        async def create_node_runs(self, body: NodeRunsCreateRequest) -> NodeRunsCreateResponse:
            self.created.append(body)
            node_run = NodeRun(
                node_run_id="child",
                node_type_ref=body.node_runs[0].node_type_ref,
                state=NodeRunState.succeeded,
                run_id=body.run_id,
            )
            return NodeRunsCreateResponse(node_runs=[node_run])

        async def get_node_run(self, node_run_id: str) -> NodeRun:
            return NodeRun(
                node_run_id=node_run_id,
                node_type_ref=NodeTypeRef(node_type_id="planner", version="1"),
                state=NodeRunState.succeeded,
                run_id="run",
            )

    planner_fixture = DevMockChatFixture(
        text="planner",
        parsed={"subtasks": [{"subtask_id": "s1", "goal": "do thing", "notes": None}], "summary": None},
    )
    arggen_fixture = DevMockChatFixture(text="arggen", parsed={})
    llm = DevMockChatModel(fixtures=[planner_fixture, arggen_fixture])
    class CompositeSelection:
        async def generate_candidate_set(self, body) -> CandidateSet:
            _ = body
            return CandidateSet(
                candidate_set_id="set",
                subtask_id="s1",
                candidates=[
                    Candidate(node_type_ref=NodeTypeRef(node_type_id="planner", version="1"), score=1.0)
                ],
            )

    coordinator = CapturingRunCoordinator()
    driver = CompositeAssignmentDriver(
        run_coordinator=coordinator,
        selection=CompositeSelection(),
        node_registry=CompositeNodeRegistry(),
        planner=Planner(llm),
        arggen=ArgGen(llm, max_retries=0),
        config=_base_config(),
        canceled=set(),
    )

    asyncio.run(driver.run(_base_request()))

    inputs = coordinator.created[0].node_runs[0].inputs
    assert inputs["goal"] == "do thing"
    assert inputs["depth"] == 1
    assert inputs["max_depth"] == 2


def test_wait_for_terminal_timeout_and_canceled() -> None:
    class FlakyRunCoordinator(RecordingRunCoordinator):
        async def get_node_run(self, node_run_id: str) -> NodeRun:
            raise RuntimeError("boom")

    driver = CompositeAssignmentDriver(
        run_coordinator=FlakyRunCoordinator(),
        selection=EmptySelection(),
        node_registry=EmptyNodeRegistry(),
        planner=Planner(DevMockChatModel(fixtures=[DevMockChatFixture(text="p", parsed={"subtasks": [], "summary": None})])),
        arggen=ArgGen(DevMockChatModel(fixtures=[DevMockChatFixture(text="a", parsed={})]), max_retries=0),
        config=_base_config(),
        canceled=set(),
    )

    result = asyncio.run(driver._wait_for_terminal("node"))
    assert result is None

    driver._canceled.add("node")
    canceled = asyncio.run(driver._wait_for_terminal("node"))
    assert canceled is None


def test_wait_for_terminal_progress() -> None:
    class ProgressRunCoordinator(RecordingRunCoordinator):
        def __init__(self) -> None:
            super().__init__()
            self.calls = 0

        async def get_node_run(self, node_run_id: str) -> NodeRun:
            self.calls += 1
            state = NodeRunState.running if self.calls == 1 else NodeRunState.succeeded
            return NodeRun(
                node_run_id=node_run_id,
                node_type_ref=NodeTypeRef(node_type_id="node", version="1"),
                state=state,
                run_id="run",
            )

    driver = CompositeAssignmentDriver(
        run_coordinator=ProgressRunCoordinator(),
        selection=EmptySelection(),
        node_registry=EmptyNodeRegistry(),
        planner=Planner(DevMockChatModel(fixtures=[DevMockChatFixture(text="p", parsed={"subtasks": [], "summary": None})])),
        arggen=ArgGen(DevMockChatModel(fixtures=[DevMockChatFixture(text="a", parsed={})]), max_retries=0),
        config=_base_config(),
        canceled=set(),
    )

    result = asyncio.run(driver._wait_for_terminal("node"))

    assert result is not None
    assert result.state == NodeRunState.succeeded



from __future__ import annotations

from arp_standard_model import Candidate, CandidateSet, Extensions, NodeTypeRef

from jarvis_composite_executor.engine.binder import choose_candidate


def test_choose_candidate_returns_first() -> None:
    candidate = Candidate(node_type_ref=NodeTypeRef(node_type_id="node", version="1"), score=1.0)
    candidate_set = CandidateSet(
        candidate_set_id="set-1",
        subtask_id="subtask",
        candidates=[candidate],
    )

    assert choose_candidate(candidate_set) == candidate


def test_choose_candidate_empty() -> None:
    candidate_set = CandidateSet(candidate_set_id="set-1", subtask_id="subtask", candidates=[])

    assert choose_candidate(candidate_set) is None


def test_choose_candidate_prefers_planner_when_flagged() -> None:
    planner = Candidate(node_type_ref=NodeTypeRef(node_type_id="jarvis.composite.planner.general", version="0.3.7"), score=1.0)
    atomic = Candidate(node_type_ref=NodeTypeRef(node_type_id="jarvis.core.echo", version="0.3.7"), score=1.0)
    candidate_set = CandidateSet(
        candidate_set_id="set-1",
        subtask_id="subtask",
        candidates=[atomic, planner],
        extensions=Extensions.model_validate(
            {
                "jarvis.selection.needs_planner": True,
                "jarvis.selection.planner_node_type_ref": {
                    "node_type_id": "jarvis.composite.planner.general",
                    "version": "0.3.7",
                },
            }
        ),
    )

    assert choose_candidate(candidate_set) == planner



from __future__ import annotations

from typing import Any

PLANNER_SYSTEM_PROMPT = (
    "You are a task planner. You decompose a goal into a small list of ordered subtasks. "
    "You will be given a JSON input with fields: task.goal, task.context, limits.max_steps, limits.depth, limits.max_depth. "
    "Your job is to output up to limits.max_steps subtasks that, when executed in order, fully achieve task.goal. "
    "Do NOT omit required work. If the goal needs more than max_steps actions at this level, create coarser subtasks that bundle work. "
    "Use depth/max_depth: if limits.depth < limits.max_depth - 1, subtasks may be higher-level bundles that will be decomposed later; "
    "When the goal contains multiple independent items, prefer subtasks that each handle one item end-to-end to minimize cross-subtask dependencies. "
    "Minimize dependencies between subtasks when feasible. If a later subtask must reuse an earlier result, say exactly what to reuse. "
    "Make subtasks explicit and self-contained (include key parameters like specific URLs or items; avoid vague references like 'the above'). "
    "Avoid duplicates and ensure coverage. "
    "Return only JSON that matches the response schema. "
)

ARGGEN_SYSTEM_PROMPT = (
    "You generate inputs for a selected action. "
    "You will be given JSON with: subtask.goal, selected_action (node_type_id, version, description), and context (including previous_steps outputs). "
    "Return ONLY a JSON object that matches the provided schema exactly. "
    "When reusing values from context.previous_steps, copy them exactly."
)

PLANNER_RESPONSE_SCHEMA: dict[str, Any] = {
    "type": "object",
    "additionalProperties": False,
    "properties": {
        "subtasks": {
            "type": "array",
            "items": {
                "type": "object",
                "additionalProperties": False,
                "properties": {
                    "subtask_id": {"type": ["string", "null"]},
                    "goal": {"type": "string"},
                    "notes": {"type": ["string", "null"]},
                },
                "required": ["subtask_id", "goal", "notes"],
            },
        },
        "summary": {"type": ["string", "null"]},
    },
    "required": ["subtasks", "summary"],
}



from __future__ import annotations

import json
import logging
from dataclasses import dataclass
from typing import Any

from arp_llm.errors import LlmError
from arp_llm.types import ChatModel, Message
from arp_standard_model import NodeType
from arp_standard_server import ArpServerError

from ..llm_assets import ARGGEN_SYSTEM_PROMPT
from ..schema_utils import find_openai_strict_schema_issues, validate_json_schema

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class ArgGenResult:
    inputs: dict[str, Any]
    provider: str
    model: str
    latency_ms: int
    usage: dict[str, Any] | None


class ArgGen:
    def __init__(self, llm: ChatModel, *, max_retries: int) -> None:
        self._llm = llm
        self._max_retries = max_retries

    async def generate_inputs(
        self,
        *,
        subtask_goal: str,
        node_type: NodeType,
        context: dict[str, Any],
    ) -> ArgGenResult:
        if node_type.input_schema is None:
            raise ArpServerError(
                code="arggen_schema_missing",
                message="NodeType is missing input_schema required for arg-gen",
                status_code=422,
                details={"node_type_id": node_type.node_type_id, "version": node_type.version},
            )
        if not isinstance(node_type.input_schema, dict):
            raise ArpServerError(
                code="arggen_schema_invalid",
                message="NodeType input_schema must be an object",
                status_code=422,
                details={"node_type_id": node_type.node_type_id, "version": node_type.version},
            )

        schema_issues = find_openai_strict_schema_issues(node_type.input_schema, limit=10)
        if schema_issues:
            logger.warning(
                "ArgGen NodeType input_schema is not OpenAI-strict (node_type_id=%s, version=%s, issues=%s)",
                node_type.node_type_id,
                node_type.version,
                schema_issues,
            )

        validation_error: dict[str, Any] | None = None
        last_error: Exception | None = None

        for attempt in range(self._max_retries + 1):
            logger.info(
                "ArgGen LLM request (attempt=%s/%s, node_type_id=%s, version=%s, goal_len=%s, has_validation_error=%s, schema_issues=%s)",
                attempt + 1,
                self._max_retries + 1,
                node_type.node_type_id,
                node_type.version,
                len(subtask_goal),
                validation_error is not None,
                len(schema_issues),
            )
            action_description = (node_type.description or "").strip() or node_type.node_type_id
            payload: dict[str, Any] = {
                "subtask": {"goal": subtask_goal},
                "selected_action": {
                    "node_type_id": node_type.node_type_id,
                    "version": node_type.version,
                    "description": action_description,
                },
                "context": context,
            }
            if validation_error is not None:
                payload["previous_validation_error"] = validation_error
            messages = [
                Message.system(ARGGEN_SYSTEM_PROMPT),
                Message.user(json.dumps(payload, sort_keys=True)),
            ]
            response = await self._llm.response(messages, response_schema=node_type.input_schema)
            if response.parsed is None:
                raise LlmError(code="parse_error", message="Arg-gen response missing parsed output")
            if not isinstance(response.parsed, dict):
                raise LlmError(code="parse_error", message="Arg-gen structured response must be a JSON object")

            parsed = response.parsed
            try:
                # Validate the raw structured output against the declared input schema.
                # (NodeType schemas are expected to be "strict" schemas where optional fields
                # are represented as nullable types rather than omitted keys.)
                validate_json_schema(parsed, node_type.input_schema)
            except Exception as exc:
                last_error = exc
                validation_error = {
                    "attempt": attempt + 1,
                    "error": str(exc),
                }
                continue

            return ArgGenResult(
                inputs=parsed,
                provider=response.provider,
                model=response.model,
                latency_ms=response.latency_ms,
                usage=response.usage,
            )

        raise ArpServerError(
            code="arggen_validation_failed",
            message="Arg-gen failed to produce valid inputs for the node schema",
            status_code=502,
            details={"error": str(last_error) if last_error else None},
        )



from __future__ import annotations

import asyncio
import logging
import time
from typing import Any

from arp_standard_model import (
    BindingDecision,
    CandidateSet,
    CandidateSetRequest,
    ConstraintEnvelope,
    Error,
    EvaluationResult,
    EvaluationStatus,
    Extensions,
    NodeKind,
    NodeRun,
    NodeRunCompleteRequest,
    NodeRunEvaluationReportRequest,
    NodeRunState,
    NodeRunTerminalState,
    NodeRunsCreateRequest,
    NodeRunCreateSpec,
    NodeType,
    SubtaskSpec,
)
from arp_standard_server import ArpServerError

from ..config import CompositeConfig
from ..utils import now
from .arggen import ArgGen
from .binder import choose_candidate
from .planner import Planner
from ..clients import NodeRegistryClientLike, RunCoordinatorClientLike, SelectionClientLike

logger = logging.getLogger(__name__)


class CompositeAssignmentDriver:
    def __init__(
        self,
        *,
        run_coordinator: RunCoordinatorClientLike,
        selection: SelectionClientLike,
        node_registry: NodeRegistryClientLike,
        planner: Planner,
        arggen: ArgGen,
        config: CompositeConfig,
        canceled: set[str],
    ) -> None:
        self._run_coordinator = run_coordinator
        self._selection = selection
        self._node_registry = node_registry
        self._planner = planner
        self._arggen = arggen
        self._config = config
        self._canceled = canceled

    async def run(self, request) -> None:
        node_run_id = request.node_run_id
        if node_run_id in self._canceled:
            await self._complete_canceled(node_run_id)
            return

        goal = _infer_goal(request.inputs)
        context = _extract_context(request.inputs)
        depth = _extract_int(request.inputs, "depth") or 0
        max_steps = _resolve_limit(
            name="max_steps",
            explicit=_extract_int(request.inputs, "max_steps"),
            constraint=_constraint_max_steps(request.constraints),
            default=self._config.max_steps_default,
        )
        max_depth = _resolve_limit(
            name="max_depth",
            explicit=_extract_int(request.inputs, "max_depth"),
            constraint=_constraint_max_depth(request.constraints),
            default=self._config.max_depth_default,
        )

        planner_result = await self._planner.plan(
            goal=goal,
            context=_planner_context(context, request.run_context),
            max_steps=max_steps,
            depth=depth,
            max_depth=max_depth,
        )

        if not planner_result.subtasks:
            await self._finalize_composite(
                node_run_id=node_run_id,
                results=[],
                planner_summary=planner_result.summary,
            )
            return

        working_context: list[dict[str, Any]] = []
        results: list[NodeRun] = []

        for subtask in planner_result.subtasks:
            if node_run_id in self._canceled:
                await self._complete_canceled(node_run_id)
                return

            candidate_set = await self._generate_candidate_set(
                request,
                subtask=subtask,
                goal=goal,
                working_context=working_context,
            )
            if (candidate := choose_candidate(candidate_set)) is None:
                raise ArpServerError(
                    code="composite_no_candidates",
                    message="Selection produced no candidates for subtask",
                    status_code=422,
                    details={"subtask_id": subtask.subtask_id},
                )

            node_type = await self._node_registry.get_node_type(
                candidate.node_type_ref.node_type_id,
                candidate.node_type_ref.version,
            )
            inputs = await self._resolve_inputs(
                request=request,
                node_type=node_type,
                candidate_set=candidate_set,
                subtask_goal=subtask.goal,
                working_context=working_context,
                depth=depth,
                max_depth=max_depth,
                max_steps=max_steps,
            )

            binding_decision = BindingDecision(
                subtask_id=subtask.subtask_id,
                candidate_set_id=candidate_set.candidate_set_id,
                chosen_node_type_ref=candidate.node_type_ref,
                rationale=candidate.rationale,
            )
            spec = _node_run_spec(
                inputs=inputs,
                candidate_set=candidate_set,
                binding_decision=binding_decision,
                constraints=request.constraints,
            )
            response = await self._run_coordinator.create_node_runs(
                NodeRunsCreateRequest(
                    run_id=request.run_id,
                    parent_node_run_id=node_run_id,
                    node_runs=[spec],
                )
            )
            if not response.node_runs:
                raise ArpServerError(
                    code="composite_no_child_runs",
                    message="Run Coordinator returned no child NodeRuns",
                    status_code=502,
                )

            child_run = response.node_runs[0]
            result = await self._wait_for_terminal(child_run.node_run_id)
            if result is None:
                if node_run_id in self._canceled:
                    await self._complete_canceled(node_run_id)
                    return
                raise ArpServerError(
                    code="composite_child_timeout",
                    message="Timed out waiting for child NodeRun completion",
                    status_code=504,
                    details={"node_run_id": child_run.node_run_id},
                )

            results.append(result)
            working_context.append(_context_step_entry(result))
            working_context = _bound_context(working_context, window=self._config.context_window)

            if result.state != NodeRunState.succeeded:
                break

        await self._finalize_composite(
            node_run_id=node_run_id,
            results=results,
            planner_summary=planner_result.summary,
        )

    async def _generate_candidate_set(
        self,
        request,
        *,
        subtask,
        goal: str,
        working_context: list[dict[str, Any]],
    ) -> CandidateSet:
        subtask_extensions_dict = _merge_extensions(
            request.extensions,
            {
                "jarvis.subtask.notes": subtask.notes,
                "jarvis.root_goal": goal,
            },
        )
        subtask_extensions = Extensions.model_validate(subtask_extensions_dict) if subtask_extensions_dict else None
        subtask_spec = SubtaskSpec(
            subtask_id=subtask.subtask_id,
            goal=subtask.goal,
            extensions=subtask_extensions,
        )
        request_extensions_dict = _merge_extensions(
            request.extensions,
            {
                "jarvis.prior_steps": working_context if working_context else None,
            },
        )
        request_extensions = Extensions.model_validate(request_extensions_dict) if request_extensions_dict else None
        return await self._selection.generate_candidate_set(
            CandidateSetRequest(
                subtask_spec=subtask_spec,
                constraints=request.constraints,
                extensions=request_extensions,
                run_context=request.run_context,
            )
        )

    async def _resolve_inputs(
        self,
        *,
        request,
        node_type: NodeType,
        candidate_set: CandidateSet,
        subtask_goal: str,
        working_context: list[dict[str, Any]],
        depth: int,
        max_depth: int,
        max_steps: int,
    ) -> dict[str, Any]:
        arg_context = _arggen_context(
            root_goal=_infer_goal(request.inputs),
            root_context=_extract_context(request.inputs),
            working_context=working_context,
            run_context=request.run_context,
            candidate_set=candidate_set,
        )
        result = await self._arggen.generate_inputs(
            subtask_goal=subtask_goal,
            node_type=node_type,
            context=arg_context,
        )
        inputs = dict(result.inputs)

        if node_type.kind == NodeKind.composite:
            _ensure_composite_inputs(
                inputs=inputs,
                goal=subtask_goal,
                context=_extract_context(request.inputs),
                depth=depth + 1,
                max_depth=max_depth,
                max_steps=max_steps,
            )

        return inputs

    async def _wait_for_terminal(self, node_run_id: str) -> NodeRun | None:
        timeout_sec = self._config.poll_timeout_sec
        interval_sec = self._config.poll_interval_sec
        deadline = time.monotonic() + timeout_sec
        while time.monotonic() <= deadline:
            if node_run_id in self._canceled:
                return None
            try:
                node_run = await self._run_coordinator.get_node_run(node_run_id)
            except Exception as exc:
                logger.warning("Failed to poll node run %s: %s", node_run_id, exc)
                await asyncio.sleep(interval_sec)
                continue
            if node_run.state in {NodeRunState.succeeded, NodeRunState.failed, NodeRunState.canceled}:
                return node_run
            await asyncio.sleep(interval_sec)
        return None

    async def _finalize_composite(
        self,
        *,
        node_run_id: str,
        results: list[NodeRun],
        planner_summary: str | None,
    ) -> None:
        evaluation_status = (
            EvaluationStatus.success
            if results and all(run.state == NodeRunState.succeeded for run in results)
            else EvaluationStatus.fail
        )
        if not results:
            evaluation_status = EvaluationStatus.success

        await self._run_coordinator.report_node_run_evaluation(
            node_run_id,
            NodeRunEvaluationReportRequest(
                evaluation_result=EvaluationResult(
                    status=evaluation_status,
                    reason_code="composite_evaluation",
                )
            ),
        )

        terminal_state = _resolve_terminal_state(results)
        error = None
        if terminal_state != NodeRunTerminalState.succeeded:
            error = Error(code="child_failed", message="One or more child NodeRuns failed")

        outputs = {
            "planner_summary": planner_summary,
            "child_node_runs": [_summarize_node_run(run) for run in results],
        }
        await self._run_coordinator.complete_node_run(
            node_run_id,
            NodeRunCompleteRequest(
                state=terminal_state,
                outputs=outputs,
                error=error,
            ),
        )

    async def _complete_canceled(self, node_run_id: str) -> None:
        await self._run_coordinator.complete_node_run(
            node_run_id,
            NodeRunCompleteRequest(state=NodeRunTerminalState.canceled),
        )


def _infer_goal(inputs: dict[str, Any]) -> str:
    if isinstance(inputs.get("goal"), str) and inputs["goal"].strip():
        return str(inputs["goal"]).strip()
    if isinstance(inputs.get("prompt"), str) and inputs["prompt"].strip():
        return str(inputs["prompt"]).strip()
    raise ArpServerError(
        code="composite_goal_missing",
        message="Composite inputs must include a goal or prompt",
        status_code=400,
    )


def _extract_context(inputs: dict[str, Any]) -> dict[str, Any]:
    context = inputs.get("context")
    if isinstance(context, dict):
        return context
    return {}


def _extract_int(inputs: dict[str, Any], name: str) -> int | None:
    value = inputs.get(name)
    if isinstance(value, int):
        return value
    if isinstance(value, str) and value.isdigit():
        return int(value)
    return None


def _resolve_limit(
    *,
    name: str,
    explicit: int | None,
    constraint: int | None,
    default: int | None,
) -> int:
    if explicit is not None:
        value = explicit
    elif constraint is not None:
        value = constraint
    else:
        value = default
    if value is None:
        raise ArpServerError(
            code="composite_limit_missing",
            message=f"{name} must be provided via inputs, constraints, or env defaults",
            status_code=400,
            details={"limit": name},
        )
    return value


def _constraint_max_steps(constraints: ConstraintEnvelope | None) -> int | None:
    if constraints and constraints.budgets:
        return constraints.budgets.max_steps
    return None


def _constraint_max_depth(constraints: ConstraintEnvelope | None) -> int | None:
    if constraints and constraints.structural:
        return constraints.structural.max_depth
    return None


def _planner_context(context: dict[str, Any], run_context) -> dict[str, Any]:
    # Keep LLM prompts focused; avoid leaking internal IDs/trace data.
    return dict(context)


def _arggen_context(
    *,
    root_goal: str,
    root_context: dict[str, Any],
    working_context: list[dict[str, Any]],
    run_context,
    candidate_set: CandidateSet,
) -> dict[str, Any]:
    return {
        "root_goal": root_goal,
        "root_context": root_context,
        "previous_steps": working_context,
    }


def _merge_extensions(base: Extensions | None, additions: dict[str, Any]) -> dict[str, Any]:
    payload: dict[str, Any] = {}
    if base is not None:
        if hasattr(base, "model_dump"):
            payload.update(base.model_dump(exclude_none=True))
        else:
            payload.update(dict(base))
    payload.update({key: value for key, value in additions.items() if value is not None})
    return payload


def _node_run_spec(
    *,
    inputs: dict[str, Any],
    candidate_set: CandidateSet,
    binding_decision: BindingDecision,
    constraints: ConstraintEnvelope | None,
):
    return NodeRunCreateSpec(
        node_type_ref=binding_decision.chosen_node_type_ref,
        inputs=inputs,
        constraints=constraints,
        candidate_set_id=candidate_set.candidate_set_id,
        binding_decision=binding_decision,
    )


def _ensure_composite_inputs(
    *,
    inputs: dict[str, Any],
    goal: str,
    context: dict[str, Any],
    depth: int,
    max_depth: int,
    max_steps: int,
) -> None:
    inputs.setdefault("goal", goal)
    inputs.setdefault("context", context)
    inputs.setdefault("depth", depth)
    inputs.setdefault("max_depth", max_depth)
    inputs.setdefault("max_steps", max_steps)


def _resolve_terminal_state(results: list[NodeRun]) -> NodeRunTerminalState:
    if any(run.state == NodeRunState.canceled for run in results):
        return NodeRunTerminalState.canceled
    if any(run.state == NodeRunState.failed for run in results):
        return NodeRunTerminalState.failed
    return NodeRunTerminalState.succeeded


def _summarize_node_run(run: NodeRun, *, subtask_id: str | None = None) -> dict[str, Any]:
    state = run.state.value if hasattr(run.state, "value") else run.state
    summary: dict[str, Any] = {
        "node_run_id": run.node_run_id,
        "node_type_ref": run.node_type_ref.model_dump(exclude_none=True),
        "state": state,
        "outputs": run.outputs,
        "output_artifacts": run.output_artifacts,
    }
    if subtask_id:
        summary["subtask_id"] = subtask_id
    return summary


def _context_step_entry(run: NodeRun) -> dict[str, Any]:
    outputs = run.outputs if run.outputs is not None else {}
    return {
        "action": run.node_type_ref.node_type_id,
        "outputs": outputs,
    }


def _bound_context(entries: list[dict[str, Any]], *, window: int | None) -> list[dict[str, Any]]:
    if window is None or window <= 0:
        return entries
    return entries[-window:]



import asyncio
import os
from typing import cast

import pytest
from arp_auth import AuthClient, AuthError
from arp_standard_server import ArpServerError

from jarvis_composite_executor import auth


class FakeToken:
    def __init__(self, access_token: str) -> None:
        self.access_token = access_token


class FakeAuthClient:
    def __init__(self, token: str = "token") -> None:
        self._token = token

    def client_credentials(self, *, audience: str | None = None, scope: str | None = None) -> FakeToken:
        _ = audience, scope
        return FakeToken(self._token)


class FakeAuthClientError(FakeAuthClient):
    def client_credentials(self, *, audience: str | None = None, scope: str | None = None) -> FakeToken:
        raise RuntimeError("boom")


def _clear_auth_env(monkeypatch: pytest.MonkeyPatch) -> None:
    for key in list(os.environ):
        if key.startswith("ARP_AUTH_"):
            monkeypatch.delenv(key, raising=False)


def test_auth_settings_default(monkeypatch: pytest.MonkeyPatch) -> None:
    _clear_auth_env(monkeypatch)

    settings = auth.auth_settings_from_env_or_dev_secure()

    assert settings.mode == "required"
    assert settings.issuer == auth.DEFAULT_DEV_KEYCLOAK_ISSUER


def test_auth_settings_from_env(monkeypatch: pytest.MonkeyPatch) -> None:
    sentinel = object()

    monkeypatch.setenv("ARP_AUTH_MODE", "required")
    monkeypatch.setattr(auth.AuthSettings, "from_env", classmethod(lambda cls: sentinel))

    assert auth.auth_settings_from_env_or_dev_secure() is sentinel


def test_auth_client_optional_none(monkeypatch: pytest.MonkeyPatch) -> None:
    _clear_auth_env(monkeypatch)
    assert auth.auth_client_from_env_optional() is None


def test_auth_client_optional_invalid(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("ARP_AUTH_CLIENT_ID", "id")
    monkeypatch.delenv("ARP_AUTH_CLIENT_SECRET", raising=False)
    with pytest.raises(RuntimeError):
        auth.auth_client_from_env_optional()


def test_auth_client_optional_success(monkeypatch: pytest.MonkeyPatch) -> None:
    sentinel = object()
    monkeypatch.setenv("ARP_AUTH_CLIENT_ID", "id")
    monkeypatch.setenv("ARP_AUTH_CLIENT_SECRET", "secret")
    monkeypatch.setattr(auth.AuthClient, "from_env", classmethod(lambda cls: sentinel))

    client = auth.auth_client_from_env_optional()

    assert client is sentinel


def test_auth_client_optional_auth_error(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setenv("ARP_AUTH_CLIENT_ID", "id")
    monkeypatch.setenv("ARP_AUTH_CLIENT_SECRET", "secret")

    def raise_auth_error(cls):
        raise AuthError("bad")

    monkeypatch.setattr(auth.AuthClient, "from_env", classmethod(raise_auth_error))

    with pytest.raises(RuntimeError):
        auth.auth_client_from_env_optional()


def test_client_credentials_token_success() -> None:
    client = FakeAuthClient(token="ok")

    result = asyncio.run(
        auth.client_credentials_token(
            auth_client=cast(AuthClient, client),
            audience=None,
            scope=None,
            service_label="svc",
        )
    )

    assert result == "ok"


def test_client_credentials_token_error() -> None:
    client = FakeAuthClientError()

    with pytest.raises(ArpServerError) as excinfo:
        asyncio.run(
            auth.client_credentials_token(
                auth_client=cast(AuthClient, client),
                audience=None,
                scope=None,
                service_label="svc",
            )
        )

    assert excinfo.value.code == "token_request_failed"



import asyncio

from arp_llm.providers.dev_mock import DevMockChatFixture, DevMockChatModel
from arp_standard_model import CompositeBeginRequest, EndpointLocator, NodeTypeRef
from jarvis_composite_executor.config import CompositeConfig
from jarvis_composite_executor.engine.arggen import ArgGen
from jarvis_composite_executor.engine.driver import CompositeAssignmentDriver
from jarvis_composite_executor.engine.planner import Planner

from .fakes import FakeNodeRegistry, FakeRunCoordinator, FakeSelection, make_candidate_set, make_node_type


def test_driver_runs_sequential_flow() -> None:
    planner_fixture = DevMockChatFixture(
        text="planner",
        parsed={"subtasks": [{"subtask_id": "s1", "goal": "do thing", "notes": None}], "summary": None},
    )
    arggen_fixture = DevMockChatFixture(text="arggen", parsed={"text": "hello"})
    llm = DevMockChatModel(fixtures=[planner_fixture, arggen_fixture])

    node_type = make_node_type(
        node_type_id="jarvis.core.echo",
        version="0.3.7",
        input_schema={
            "type": "object",
            "properties": {"text": {"type": "string"}},
            "required": ["text"],
            "additionalProperties": False,
        },
    )
    candidate_ref = NodeTypeRef(node_type_id="jarvis.core.echo", version="0.3.7")
    selection = FakeSelection(make_candidate_set(candidate=candidate_ref))
    node_registry = FakeNodeRegistry(node_type)
    run_coordinator = FakeRunCoordinator()

    config = CompositeConfig(
        max_steps_default=3,
        max_depth_default=2,
        poll_timeout_sec=1.0,
        poll_interval_sec=0.01,
        arggen_max_retries=0,
        context_window=None,
        max_concurrency=None,
    )

    driver = CompositeAssignmentDriver(
        run_coordinator=run_coordinator,
        selection=selection,
        node_registry=node_registry,
        planner=Planner(llm),
        arggen=ArgGen(llm, max_retries=0),
        config=config,
        canceled=set(),
    )

    request = CompositeBeginRequest(
        run_id="run_1",
        node_run_id="node_run_1",
        node_type_ref=NodeTypeRef(node_type_id="jarvis.composite.planner", version="0.3.7"),
        inputs={"goal": "test"},
        coordinator_endpoint=EndpointLocator.model_validate("http://127.0.0.1:8081"),
    )

    asyncio.run(driver.run(request))

    assert len(run_coordinator.created) == 1
    assert len(run_coordinator.completed) == 1



from __future__ import annotations

import logging
import os

from arp_llm.settings import load_chat_model_from_env

from .auth import auth_client_from_env_optional, auth_settings_from_env_or_dev_secure
from .clients import NodeRegistryGatewayClient, SelectionGatewayClient
from .config import load_composite_config_from_env
from .executor import CompositeExecutor
from .utils import normalize_base_url

logger = logging.getLogger(__name__)


def create_app():
    selection_url = _require_url("JARVIS_SELECTION_URL", fallback="ARP_SELECTION_URL")
    node_registry_url = _require_url("JARVIS_NODE_REGISTRY_URL", fallback="ARP_NODE_REGISTRY_URL")
    logger.info(
        "Composite Executor upstreams (selection_url=%s, node_registry_url=%s)",
        selection_url,
        node_registry_url,
    )

    if (auth_client := auth_client_from_env_optional()) is None:
        raise RuntimeError("ARP_AUTH_CLIENT_ID and ARP_AUTH_CLIENT_SECRET are required for outbound auth.")

    selection_service = SelectionGatewayClient(
        base_url=selection_url,
        auth_client=auth_client,
        audience=_audience_from_env("JARVIS_SELECTION_AUDIENCE"),
    )
    node_registry = NodeRegistryGatewayClient(
        base_url=node_registry_url,
        auth_client=auth_client,
        audience=_audience_from_env("JARVIS_NODE_REGISTRY_AUDIENCE"),
    )

    llm = load_chat_model_from_env()
    config = load_composite_config_from_env()
    logger.info(
        "Composite Executor config (max_steps_default=%s, max_depth_default=%s, poll_timeout_sec=%s, poll_interval_sec=%s, arggen_max_retries=%s, context_window=%s, max_concurrency=%s)",
        config.max_steps_default,
        config.max_depth_default,
        config.poll_timeout_sec,
        config.poll_interval_sec,
        config.arggen_max_retries,
        config.context_window,
        config.max_concurrency,
    )

    auth_settings = auth_settings_from_env_or_dev_secure()
    logger.info(
        "Composite Executor auth settings (mode=%s, issuer=%s)",
        getattr(auth_settings, "mode", None),
        getattr(auth_settings, "issuer", None),
    )
    return CompositeExecutor(
        selection_service=selection_service,
        node_registry=node_registry,
        llm=llm,
        config=config,
    ).create_app(
        title="JARVIS Composite Executor",
        auth_settings=auth_settings,
    )


def _require_url(name: str, *, fallback: str | None = None) -> str:
    value = (os.environ.get(name) or "").strip()
    if not value and fallback:
        value = (os.environ.get(fallback) or "").strip()
    if not value:
        raise RuntimeError(f"{name} is required to start the Composite Executor")
    return normalize_base_url(value)


def _audience_from_env(name: str) -> str | None:
    value = (os.environ.get(name) or "").strip()
    if value:
        return value
    return None


app = create_app()



from __future__ import annotations

import json
import logging
from dataclasses import dataclass
from typing import Any

from arp_llm.errors import LlmError
from arp_llm.types import ChatModel, Message
from arp_standard_server import ArpServerError

from ..llm_assets import PLANNER_RESPONSE_SCHEMA, PLANNER_SYSTEM_PROMPT
from ..schema_utils import find_openai_strict_schema_issues

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class PlannedSubtask:
    subtask_id: str
    goal: str
    notes: str | None


@dataclass(frozen=True)
class PlannerResult:
    subtasks: list[PlannedSubtask]
    summary: str | None
    provider: str
    model: str
    latency_ms: int
    usage: dict[str, Any] | None


class Planner:
    def __init__(self, llm: ChatModel) -> None:
        self._llm = llm

    async def plan(
        self,
        *,
        goal: str,
        context: dict[str, Any],
        max_steps: int,
        depth: int,
        max_depth: int,
    ) -> PlannerResult:
        if max_steps < 1:
            raise ArpServerError(
                code="planner_invalid_max_steps",
                message="max_steps must be >= 1 for planning",
                status_code=400,
                details={"max_steps": max_steps},
            )
        if max_depth < 0:
            raise ArpServerError(
                code="planner_invalid_max_depth",
                message="max_depth must be >= 0 for planning",
                status_code=400,
                details={"max_depth": max_depth},
            )
        if depth >= max_depth:
            raise ArpServerError(
                code="planner_depth_exceeded",
                message="Composite depth exceeds max_depth",
                status_code=400,
                details={"depth": depth, "max_depth": max_depth},
            )

        context_payload: dict[str, Any] = dict(context)
        context_payload.pop("run_context", None)

        payload: dict[str, Any] = {
            "task": {
                "goal": goal,
                "context": context_payload,
            },
            "limits": {
                "max_steps": max_steps,
                "depth": depth,
                "max_depth": max_depth,
            },
        }
        messages = [
            Message.system(PLANNER_SYSTEM_PROMPT),
            Message.user(json.dumps(payload, sort_keys=True)),
        ]

        schema_issues = find_openai_strict_schema_issues(PLANNER_RESPONSE_SCHEMA, limit=5)
        if schema_issues:
            logger.warning("Planner response schema is not OpenAI-strict (issues=%s)", schema_issues)
        logger.info(
            "Planner LLM request (goal_len=%s, max_steps=%s, depth=%s/%s, schema_issues=%s)",
            len(goal),
            max_steps,
            depth,
            max_depth,
            len(schema_issues),
        )
        response = await self._llm.response(messages, response_schema=PLANNER_RESPONSE_SCHEMA)
        if response.parsed is None:
            raise LlmError(code="parse_error", message="Planner response missing parsed output")
        parsed = response.parsed

        raw_subtasks = parsed.get("subtasks")
        if not isinstance(raw_subtasks, list):
            raise ArpServerError(
                code="planner_invalid_output",
                message="Planner response missing subtasks array",
                status_code=502,
            )

        subtasks = self._normalize_subtasks(raw_subtasks, max_steps=max_steps)
        summary = parsed.get("summary")
        summary_value = summary if isinstance(summary, str) else None

        return PlannerResult(
            subtasks=subtasks,
            summary=summary_value,
            provider=response.provider,
            model=response.model,
            latency_ms=response.latency_ms,
            usage=response.usage,
        )

    def _normalize_subtasks(
        self, raw_subtasks: list[Any], *, max_steps: int
    ) -> list[PlannedSubtask]:
        normalized: list[PlannedSubtask] = []
        for index, item in enumerate(raw_subtasks):
            if len(normalized) >= max_steps:
                break
            if not isinstance(item, dict):
                raise ArpServerError(
                    code="planner_invalid_subtask",
                    message="Planner subtask must be an object",
                    status_code=502,
                    details={"index": index},
                )
            goal = item.get("goal")
            notes = item.get("notes")
            if not isinstance(goal, str) or not goal.strip():
                raise ArpServerError(
                    code="planner_invalid_subtask",
                    message="Planner subtask goal must be a non-empty string",
                    status_code=502,
                    details={"index": index},
                )
            subtask_id = f"S{len(normalized) + 1}"
            note_value = notes if isinstance(notes, str) else None
            normalized.append(PlannedSubtask(subtask_id=subtask_id, goal=goal, notes=note_value))

        return normalized



#!/usr/bin/env python3
from __future__ import annotations

import argparse
import json
from pathlib import Path
import sys
from urllib.error import HTTPError, URLError
from urllib.request import Request, urlopen

DEFAULT_REQUEST_PATH = Path(__file__).with_name("request.json")


def _request(method: str, url: str, body: dict | None, headers: dict[str, str]) -> object:
    data = None
    if body is not None:
        data = json.dumps(body).encode("utf-8")
    req = Request(url, data=data, headers=headers, method=method)
    try:
        with urlopen(req) as resp:
            payload = resp.read().decode("utf-8")
    except HTTPError as exc:
        payload = exc.read().decode("utf-8")
        raise SystemExit(f"HTTP {exc.code} {exc.reason}: {payload}") from exc
    except URLError as exc:
        raise SystemExit(f"Request failed: {exc.reason}") from exc
    if not payload:
        return None
    try:
        return json.loads(payload)
    except json.JSONDecodeError:
        return payload


def _load_request(path: Path) -> dict:
    payload = json.loads(path.read_text(encoding="utf-8"))
    if not isinstance(payload, dict):
        raise SystemExit("Request JSON must be an object.")
    return payload


def main() -> int:
    parser = argparse.ArgumentParser(description="Send a composite begin request from a JSON file.")
    parser.add_argument("--host", default="127.0.0.1")
    parser.add_argument("--port", default="8083")
    parser.add_argument("--request", default=str(DEFAULT_REQUEST_PATH), help="Path to a JSON request body.")
    args = parser.parse_args()

    request_path = Path(args.request)
    body = _load_request(request_path)

    base_url = f"http://{args.host}:{args.port}/v1"
    headers = {"Content-Type": "application/json"}

    result = _request("POST", f"{base_url}/composite-node-runs:begin", body, headers)
    print(json.dumps({"result": result}, indent=2, sort_keys=True))
    return 0


if __name__ == "__main__":
    sys.exit(main())




import asyncio

from arp_llm.providers.dev_mock import DevMockChatFixture, DevMockChatModel
from jarvis_composite_executor.engine.planner import Planner


def test_planner_truncates_and_assigns_ids() -> None:
    fixture = DevMockChatFixture(
        text="planner",
        parsed={
            "subtasks": [
                {"subtask_id": "", "goal": "step one", "notes": None},
                {"subtask_id": "step-2", "goal": "step two", "notes": None},
            ],
            "summary": "two steps",
        },
    )
    planner = Planner(DevMockChatModel(fixtures=[fixture]))

    result = asyncio.run(
        planner.plan(
            goal="root goal",
            context={},
            max_steps=1,
            depth=0,
            max_depth=2,
        )
    )

    assert len(result.subtasks) == 1
    assert result.subtasks[0].subtask_id == "S1"



from __future__ import annotations

import copy
from typing import Any, Mapping


def normalize_schema_for_llm(schema: Mapping[str, Any]) -> dict[str, Any]:
    normalized = copy.deepcopy(dict(schema))
    return _normalize_schema(normalized)


def find_openai_strict_schema_issues(schema: Mapping[str, Any], *, limit: int = 25) -> list[dict[str, Any]]:
    """
    Best-effort validator for OpenAI "strict" structured output schemas.

    OpenAI requires every object schema to explicitly set `"additionalProperties": false`, and
    for `"required"` to contain all keys in `"properties"` (nullable types are used to represent
    optionality).

    Returns a bounded list of issues suitable for logging.
    """

    issues: list[dict[str, Any]] = []

    def visit(value: Any, path: list[Any]) -> None:
        if len(issues) >= limit:
            return
        if isinstance(value, dict):
            schema_type = value.get("type")
            is_object = schema_type == "object" or (isinstance(schema_type, list) and "object" in schema_type)

            if is_object:
                ap = value.get("additionalProperties", None)
                if ap is not False:
                    issues.append(
                        {
                            "path": _format_schema_path(path),
                            "issue": "additionalProperties_must_be_false",
                            "value": ap,
                        }
                    )
                    if len(issues) >= limit:
                        return

                props = value.get("properties")
                if props is None or not isinstance(props, dict):
                    issues.append(
                        {
                            "path": _format_schema_path(path + ["properties"]),
                            "issue": "properties_must_be_object",
                            "value": None if props is None else type(props).__name__,
                        }
                    )
                    if len(issues) >= limit:
                        return
                    props = {}

                req = value.get("required")
                if req is None or not isinstance(req, list):
                    issues.append(
                        {
                            "path": _format_schema_path(path + ["required"]),
                            "issue": "required_must_be_array",
                            "value": None if req is None else type(req).__name__,
                        }
                    )
                    if len(issues) >= limit:
                        return
                    req = []

                missing = [key for key in props.keys() if key not in req]
                if missing:
                    issues.append(
                        {
                            "path": _format_schema_path(path + ["required"]),
                            "issue": "required_must_include_all_properties",
                            "value": missing,
                        }
                    )
                    if len(issues) >= limit:
                        return

            for key, item in value.items():
                if key in ("$defs", "definitions") and isinstance(item, dict):
                    for def_name, def_schema in item.items():
                        visit(def_schema, path + [key, def_name])
                    continue
                if key in ("anyOf", "oneOf", "allOf") and isinstance(item, list):
                    for idx, branch in enumerate(item):
                        visit(branch, path + [key, idx])
                    continue
                if key == "properties" and isinstance(item, dict):
                    for prop_name, prop_schema in item.items():
                        visit(prop_schema, path + [key, prop_name])
                    continue
                if key == "items":
                    visit(item, path + [key])
                    continue
        elif isinstance(value, list):
            for idx, item in enumerate(value):
                visit(item, path + [idx])

    visit(dict(schema), [])
    return issues


def _format_schema_path(path: list[Any]) -> str:
    if not path:
        return "$"
    parts: list[str] = ["$"]
    for item in path:
        if isinstance(item, int):
            parts.append(f"[{item}]")
        else:
            parts.append(f".{item}")
    return "".join(parts)


def validate_json_schema(instance: Mapping[str, Any], schema: Mapping[str, Any]) -> None:
    import jsonschema

    jsonschema.validate(instance=instance, schema=dict(schema))


def _normalize_schema(schema: dict[str, Any]) -> dict[str, Any]:
    if "anyOf" in schema and isinstance(schema["anyOf"], list):
        schema["anyOf"] = [_normalize_schema(_as_dict(item)) for item in schema["anyOf"]]
    if "oneOf" in schema and isinstance(schema["oneOf"], list):
        schema["oneOf"] = [_normalize_schema(_as_dict(item)) for item in schema["oneOf"]]

    schema_type = schema.get("type")
    if schema_type == "object":
        properties = schema.get("properties")
        if not isinstance(properties, dict):
            schema["properties"] = {}
            schema["required"] = []
            schema["additionalProperties"] = False
            return schema

        original_required = set(schema.get("required") or [])
        required = list(properties.keys())
        schema["required"] = required
        schema["additionalProperties"] = False
        for key, prop_schema in list(properties.items()):
            normalized = _normalize_schema(_as_dict(prop_schema))
            if key not in original_required:
                normalized = _allow_null(normalized)
            properties[key] = normalized
        return schema

    if schema_type == "array":
        items = schema.get("items")
        if isinstance(items, dict):
            schema["items"] = _normalize_schema(items)
        return schema

    return schema


def _allow_null(schema: dict[str, Any]) -> dict[str, Any]:
    if "type" in schema:
        schema_type = schema["type"]
        if isinstance(schema_type, list):
            if "null" not in schema_type:
                schema["type"] = list(schema_type) + ["null"]
        elif isinstance(schema_type, str):
            if schema_type != "null":
                schema["type"] = [schema_type, "null"]
        return schema

    any_of = schema.get("anyOf")
    if isinstance(any_of, list):
        normalized = [_as_dict(item) for item in any_of]
        if not any(_allows_null(item) for item in normalized):
            normalized.append({"type": "null"})
        schema["anyOf"] = normalized
        return schema

    schema["anyOf"] = [schema, {"type": "null"}]
    return schema


def _allows_null(schema: dict[str, Any]) -> bool:
    schema_type = schema.get("type")
    if schema_type == "null":
        return True
    if isinstance(schema_type, list) and "null" in schema_type:
        return True
    if "anyOf" in schema and isinstance(schema["anyOf"], list):
        return any(_allows_null(_as_dict(item)) for item in schema["anyOf"])
    if "oneOf" in schema and isinstance(schema["oneOf"], list):
        return any(_allows_null(_as_dict(item)) for item in schema["oneOf"])
    return False


def _as_dict(value: Any) -> dict[str, Any]:
    if isinstance(value, dict):
        return value
    if hasattr(value, "model_dump"):
        return dict(value.model_dump(exclude_none=True))
    return dict(value)



from __future__ import annotations

from datetime import datetime, timezone


def now() -> datetime:
    return datetime.now(timezone.utc)


def normalize_base_url(url: str) -> str:
    normalized = url.rstrip("/")
    if normalized.endswith("/v1"):
        normalized = normalized[:-3]
    return normalized



from __future__ import annotations

import asyncio
import logging
from collections.abc import Awaitable, Callable

from arp_llm.types import ChatModel
from arp_standard_model import (
    CompositeBeginRequest,
    CompositeBeginResponse,
    CompositeExecutorBeginCompositeNodeRunRequest,
    CompositeExecutorCancelCompositeNodeRunRequest,
    CompositeExecutorHealthRequest,
    CompositeExecutorVersionRequest,
    Error,
    Health,
    NodeRunCompleteRequest,
    NodeRunTerminalState,
    Status,
    VersionInfo,
)
from arp_standard_server import ArpServerError
from arp_standard_server.composite_executor import BaseCompositeExecutorServer

from . import __version__
from .clients import (
    NodeRegistryClientLike,
    RunCoordinatorClientLike,
    RunCoordinatorGatewayClient,
    SelectionClientLike,
)
from .config import CompositeConfig
from .engine import ArgGen, CompositeAssignmentDriver, Planner
from .utils import normalize_base_url, now

logger = logging.getLogger(__name__)


class CompositeExecutor(BaseCompositeExecutorServer):
    """Composite execution surface; plug your composite engine here."""

    # Core method - API surface and main extension points
    def __init__(
        self,
        *,
        service_name: str = "arp-jarvis-composite-executor",
        service_version: str = __version__,
        selection_service: SelectionClientLike | None = None,
        node_registry: NodeRegistryClientLike | None = None,
        llm: ChatModel | None = None,
        config: CompositeConfig | None = None,
        run_coordinator_factory: Callable[[CompositeBeginRequest], RunCoordinatorClientLike] | None = None,
    ) -> None:
        """
        Not part of ARP spec; required to construct the executor.

        Args:
          - service_name: Name exposed by /v1/version.
          - service_version: Version exposed by /v1/version.
          - run_coordinator: Optional wrapper for Run Coordinator calls.
          - selection_service: Optional wrapper for Selection Service calls.

        Potential modifications:
          - Inject your composite planner/executor implementation.
          - Add background task orchestration or queues.
        """
        self._service_name = service_name
        self._service_version = service_version
        self._selection_service = selection_service
        self._node_registry = node_registry
        self._llm = llm
        self._config = config
        self._run_coordinator_factory = run_coordinator_factory
        self._tasks: dict[str, asyncio.Task[None]] = {}
        self._canceled: set[str] = set()
        self._semaphore = None
        if config and config.max_concurrency and config.max_concurrency > 0:
            self._semaphore = asyncio.Semaphore(config.max_concurrency)

    # Core methods - Composite Executor API implementations
    async def health(self, request: CompositeExecutorHealthRequest) -> Health:
        """
        Mandatory: Required by the ARP Composite Executor API.

        Args:
          - request: CompositeExecutorHealthRequest (unused).
        """
        _ = request
        return Health(status=Status.ok, time=now())

    async def version(self, request: CompositeExecutorVersionRequest) -> VersionInfo:
        """
        Mandatory: Required by the ARP Composite Executor API.

        Args:
          - request: CompositeExecutorVersionRequest (unused).
        """
        _ = request
        return VersionInfo(
            service_name=self._service_name,
            service_version=self._service_version,
            supported_api_versions=["v1"],
        )

    async def begin_composite_node_run(
        self, request: CompositeExecutorBeginCompositeNodeRunRequest
    ) -> CompositeBeginResponse:
        """
        Mandatory: Required by the ARP Composite Executor API.

        Args:
          - request: CompositeExecutorBeginCompositeNodeRunRequest with assignment info.

        Potential modifications:
          - Kick off background execution and return accepted=true.
          - Validate constraints before accepting.
        """
        self._ensure_ready()
        logger.info(
            "Composite assignment received (run_id=%s, node_run_id=%s, node_type_id=%s)",
            request.body.run_id,
            request.body.node_run_id,
            request.body.node_type_ref.node_type_id,
        )
        self._start_assignment(request.body)
        return CompositeBeginResponse(
            accepted=True,
            message="Composite assignment accepted (JARVIS).",
        )

    async def cancel_composite_node_run(self, request: CompositeExecutorCancelCompositeNodeRunRequest) -> None:
        """
        Mandatory: Required by the ARP Composite Executor API.

        Args:
          - request: CompositeExecutorCancelCompositeNodeRunRequest with node_run_id.

        Potential modifications:
          - Add cooperative cancellation to your composite executor implementation.
        """
        node_run_id = request.params.node_run_id
        self._canceled.add(node_run_id)
        task = self._tasks.pop(node_run_id, None)
        if task is not None:
            task.cancel()
            logger.info("Composite cancel requested (node_run_id=%s)", node_run_id)
        return None

    def _start_assignment(self, request: CompositeBeginRequest) -> None:
        node_run_id = request.node_run_id
        if node_run_id in self._tasks:
            return
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            asyncio.run(self._drive_assignment(request))
            return
        task = loop.create_task(self._drive_assignment(request))
        task.add_done_callback(lambda _: self._tasks.pop(node_run_id, None))
        self._tasks[node_run_id] = task

    async def _drive_assignment(self, request: CompositeBeginRequest) -> None:
        node_run_id = request.node_run_id
        if node_run_id in self._canceled:
            return
        selection = self._selection_service
        node_registry = self._node_registry
        llm = self._llm
        config = self._config
        if selection is None or node_registry is None or llm is None or config is None:
            logger.error("Composite assignment missing dependencies (node_run_id=%s)", node_run_id)
            return
        run_coordinator = (
            self._run_coordinator_factory(request)
            if self._run_coordinator_factory
            else RunCoordinatorGatewayClient(
                base_url=normalize_base_url(str(request.coordinator_endpoint.root)),
                bearer_token=request.assignment_token,
            )
        )
        try:
            driver = CompositeAssignmentDriver(
                run_coordinator=run_coordinator,
                selection=selection,
                node_registry=node_registry,
                planner=Planner(llm),
                arggen=ArgGen(llm, max_retries=config.arggen_max_retries),
                config=config,
                canceled=self._canceled,
            )
            await _maybe_await(self._semaphore, driver.run(request))
            logger.info("Composite assignment completed (node_run_id=%s)", node_run_id)
        except asyncio.CancelledError:
            try:
                await run_coordinator.complete_node_run(
                    node_run_id,
                    NodeRunCompleteRequest(state=NodeRunTerminalState.canceled),
                )
            except Exception:
                logger.exception("Unable to report composite cancellation")
            return
        except ArpServerError as exc:
            logger.exception("Composite assignment failed", exc_info=exc)
            await _report_failure(run_coordinator, node_run_id, exc)
        except Exception as exc:
            logger.exception("Composite assignment failed", exc_info=exc)
            await _report_failure(
                run_coordinator,
                node_run_id,
                ArpServerError(code="composite_error", message=str(exc), status_code=502),
            )

    def _ensure_ready(self) -> None:
        if self._selection_service is None:
            logger.error("Selection Service is not configured for Composite Executor")
            raise ArpServerError(
                code="selection_service_missing",
                message="Selection Service is required for composite execution",
                status_code=503,
            )
        if self._node_registry is None:
            logger.error("Node Registry is not configured for Composite Executor")
            raise ArpServerError(
                code="node_registry_missing",
                message="Node Registry is required for composite execution",
                status_code=503,
            )
        if self._llm is None:
            logger.error("LLM is not configured for Composite Executor")
            raise ArpServerError(
                code="composite_llm_missing",
                message="Composite Executor requires an LLM for planning and arg-gen",
                status_code=503,
            )
        if self._config is None:
            logger.error("Composite Executor config is not configured")
            raise ArpServerError(
                code="composite_config_missing",
                message="Composite Executor configuration is required",
                status_code=503,
            )


async def _report_failure(
    run_coordinator: RunCoordinatorClientLike,
    node_run_id: str,
    exc: ArpServerError,
) -> None:
    try:
        await run_coordinator.complete_node_run(
            node_run_id,
            NodeRunCompleteRequest(
                state=NodeRunTerminalState.failed,
                error=Error(code=exc.code, message=exc.message, details=exc.details),
            ),
        )
    except Exception:
        logger.exception("Unable to report composite failure")


async def _maybe_await(
    semaphore: asyncio.Semaphore | None,
    coro: Awaitable[None],
) -> None:
    if semaphore is None:
        await coro
        return
    async with semaphore:
        await coro
